      PROGRAM TMKLM
      IMPLICIT INTEGER(A-Z)
      DOUBLE PRECISION TIMEA,TIMEB,TLIMIT,VAF,ZMIN,ZBEST,Z,DMIN,
     1                 DCOM,CMIN,ASUM,ASSE
      REAL S1
      DOUBLE PRECISION, ALLOCATABLE :: A(:,:) 
      DOUBLE PRECISION, ALLOCATABLE :: V(:,:)
      DOUBLE PRECISION, ALLOCATABLE :: C(:,:) 
      DOUBLE PRECISION, ALLOCATABLE :: D(:,:)
      DOUBLE PRECISION, ALLOCATABLE :: CENTR(:,:)
      DOUBLE PRECISION, ALLOCATABLE :: CENTC(:,:)
      INTEGER, ALLOCATABLE :: NR(:)
      INTEGER, ALLOCATABLE :: NC(:)
      INTEGER, ALLOCATABLE :: CMEM(:)
      INTEGER, ALLOCATABLE :: RMEM(:)
      INTEGER, ALLOCATABLE :: CBEST(:)
      INTEGER, ALLOCATABLE :: RBEST(:)
C
C #########################################################################
C  JANUARY 2019 - FORTRAN IMPLEMENTATION OF A MULTISTART TWO-MODE KL-MEANS
C                 PARTITIONING ALGORITHM (WCSS CRITERION)
C                 BLOCK PLACEMENTS ARE ASSUMED **UNKNOWN**
C  INPUT FILES:
C      1) PARSET -  SINGLE ROW PARAMETER FILE CONTAINING THE NUMBER OF 
C                   ROW OBJECTS (RO), NUMBER OF COLUMN OBJECTS (CO),
C                   NUMBER OF ROW CLUSTERS (RC), NUMBER OF COLUMN
C                   CLUSTERS (CC), AND TIME LIMIT (TL)
C      2) DATAFIL - AN RO-BY-CO REAL-VALUED MATRIX, FREE INPUT, READ
C                   BY ROW
C  OUTPUT FILES:
C      3) OUTPT   - A SUMMARY OUTPUT FILE THAT REPORTS THE VARIATION
C                   ACCOUNTED FOR (VAF) AND NUMBER OF RESTARTS
C      4) ROWPART - CLUSTER ASSIGNMENTS OF ROW OBJECTS
C      5) COLPART - CLUSTER ASSIGNMENTS OF COLUMN OBJECTS
C #######################################################################
C     
      OPEN(1,FILE='DATAFIL')
      OPEN(2,FILE='OUTPT')
      OPEN(3,FILE='ROWPART')
      OPEN(4,FILE='COLPART')
      OPEN(5,FILE='PARSET')
      READ(5,*) RO,CO,RC,CC,TLIMIT
      ALLOCATE (A(RO,CO),V(RC,CC),C(RO,RC),D(CO,CC))
      ALLOCATE (CENTR(RC,CO),CENTC(RO,CC))
      ALLOCATE (NR(RC),NC(CC),CMEM(CO),RMEM(RO),CBEST(CO),RBEST(RO))
      READ(1,*) ((A(I,J),J=1,CO),I=1,RO)     
      CALL CPU_TIME(TIMEA)
      NREPS = 0
      NUMOPT = 1
      UNIQUE = 1
      ZBEST = 999999999.0D0
      ASUM = 0.0D0
      ASSE = 0.0D0
      DO I = 1,RO
        DO J = 1,CO
          ASUM = ASUM + A(I,J)
        END DO
      END DO
      ASUM = ASUM/DFLOAT(RO*CO)
      DO I = 1,RO
        DO J = 1,CO
          ASSE = ASSE + (A(I,J)-ASUM)**2
        END DO
      END DO
 1000 NREPS = NREPS + 1
        DO I = 1,RC
          NR(I) = 0
          DO J = 1,CC
            NC(J) = 0
            V(I,J) = 0.0D0
          END DO
        END DO
        DO I = 1,RO
          RMEM(I) = 0
        END DO
        DO J = 1,CO
          CMEM(J) = 0
        END DO
C
C  ###############################################################
C              GREEDY INITIALIZATION PROCEDURE - PART 1:
C
C    RANDOMLY SELECT K1 OF THE ROW OBJECTS TO PUT ONE OBJECT IN
C    EACH OF THE FIRST K1 ROW CLUSTERS.  THEN, DO THE SAME FOR
C    TO OBTAIN K2 COLUMN OBJECTS TO PUT ONE OBJECT IN EACH OF
C    THE FIRST K2 COLUMN CLUSTERS.  THEN COMPUTE THE OBJECTIVE
C    VALUE.
C  ################################################################
C
        DO K = 1,RC
 98       CALL RANDOM_NUMBER(S1)
          ISEL = S1 * FLOAT(RO) + 1.
          IF(ISEL.GT.RO) ISEL = RO
          IF(RMEM(ISEL).GT.0) GO TO 98
          RMEM(ISEL) = K
          NR(K) = NR(K) + 1
          DO J = 1,CO
            CENTR(K,J) = A(ISEL,J)
          END DO
        END DO
        DO L = 1,CC
 97       CALL RANDOM_NUMBER(S1)
          JSEL = S1 * FLOAT(CO) + 1.
          IF(JSEL.GT.CO) JSEL = CO
          IF(CMEM(JSEL).GT.0) GO TO 97
          CMEM(JSEL) = L
          NC(L) = NC(L) + 1
          DO I = 1,RO
            CENTC(I,L) = A(I,JSEL)
          END DO
        END DO
        DO 300 I = 1,RO
          IF(RMEM(I).NE.0) GO TO 300
          DMIN = 9999999
          DO K = 1,RC
            DCOM = 0
            DO J = 1,CO
              DCOM = DCOM + (A(I,J)-CENTR(K,J))**2
            END DO
            IF(DCOM.LT.DMIN) THEN
              DMIN = DCOM
              KSEL = K
            END IF
          END DO
          RMEM(I) = KSEL
          NR(KSEL) = NR(KSEL) + 1
 300    CONTINUE
        DO 301 J = 1,CO
          IF(CMEM(J).NE.0) GO TO 301
          DMIN = 9999999
          DO K = 1,CC
            DCOM = 0
            DO I = 1,RO
              DCOM = DCOM + (A(I,J)-CENTC(I,K))**2
            END DO
            IF(DCOM.LT.DMIN) THEN
              DMIN = DCOM
              KSEL = K
            END IF
          END DO
          CMEM(J) = KSEL
          NC(KSEL) = NC(KSEL) + 1
 301    CONTINUE
C
                              ! COMPUTE CENTROIDS
        Z = 0.0D0
        DO I = 1,RO
          KR = RMEM(I)
          DO J = 1,CO
            KC = CMEM(J)
            V(KR,KC) = V(KR,KC)+A(I,J)
          END DO
        END DO
        DO K = 1,RC
          DO L = 1,CC
            IF(NR(K).GT.0.AND.NC(L).GT.0) THEN
              V(K,L) = V(K,L)/DFLOAT(NR(K)*NC(L))
            END IF
          END DO
        END DO
        DO I = 1,RO
          K = RMEM(I)
          DO J = 1,CO
            L = CMEM(J)
            Z = Z + (A(I,J)-V(K,L))**2
          END DO
        END DO
        ZMIN = Z
C
 450    TRIG = 0
        DO I = 1,RO
          DO K = 1,RC
            C(I,K) = 0.0D0
          END DO
        END DO
        DO 451 I = 1,RO
          DO 452 K = 1,RC
            DO 453 J = 1,CO
              L = CMEM(J)
              C(I,K) = C(I,K) + (A(I,J) - V(K,L))**2
 453        CONTINUE
 452      CONTINUE
 451    CONTINUE
        DO K = 1,RC
          NR(K) = 0
        END DO
        DO 463 I = 1,RO
          CMIN = 9.9D+12
          DO 464 K = 1,RC
            IF(C(I,K).LT.CMIN) THEN
              CMIN = C(I,K)
              KSEL = K
            END IF
 464      CONTINUE
          RMEM(I) = KSEL
          NR(KSEL) = NR(KSEL) + 1
 463    CONTINUE
C
        DO K = 1,RC
          DO L = 1,CC
            V(K,L) = 0.0D0
          END DO
        END DO
        Z = 0.0D0
        DO I = 1,RO
          KR = RMEM(I)
          DO J = 1,CO
            KC = CMEM(J)
            V(KR,KC) = V(KR,KC)+A(I,J)
          END DO
        END DO
        DO K = 1,RC
          DO L = 1,CC
            IF(NR(K).GT.0.AND.NC(L).GT.0) THEN
              V(K,L) = V(K,L)/DFLOAT(NR(K)*NC(L))
            END IF
          END DO
        END DO
        DO I = 1,RO
          K = RMEM(I)
          DO J = 1,CO
            L = CMEM(J)
            Z = Z + (A(I,J)-V(K,L))**2
          END DO
        END DO
C
        DO J = 1,CO
          DO L = 1,CC
            D(J,L) = 0.0D0
          END DO
        END DO
        DO 455 J = 1,CO
          DO 456 L = 1,CC
            DO 457 I= 1,RO
              K = RMEM(I)
              D(J,L) = D(J,L) + (A(I,J) - V(K,L))**2
 457        CONTINUE
 456      CONTINUE
 455    CONTINUE
        DO L = 1,CC
          NC(L) = 0
        END DO
        DO 467 J = 1,CO
          CMIN = 9.9D+12
          DO 468 L = 1,CC
            IF(D(J,L).LT.CMIN) THEN
              CMIN = D(J,L)
              LSEL = L
            END IF
 468      CONTINUE
          CMEM(J) = LSEL
          NC(LSEL) = NC(LSEL) + 1
 467    CONTINUE
C
        DO K = 1,RC
          DO L = 1,CC
            V(K,L) = 0.0D0
          END DO
        END DO
        Z = 0.0D0
        DO I = 1,RO
          KR = RMEM(I)
          DO J = 1,CO
            KC = CMEM(J)
            V(KR,KC) = V(KR,KC)+A(I,J)
          END DO
        END DO
        DO K = 1,RC
          DO L = 1,CC
            IF(NR(K).GT.0.AND.NC(L).GT.0) THEN
              V(K,L) = V(K,L)/DFLOAT(NR(K)*NC(L))
            END IF
          END DO
        END DO
        DO I = 1,RO
          K = RMEM(I)
          DO J = 1,CO
            L = CMEM(J)
            Z = Z + (A(I,J)-V(K,L))**2
          END DO
        END DO
C
        IF(Z.LT.ZMIN-.000001) THEN
          ZMIN = Z
          GO TO 450
        END IF
C
        IF(ZMIN.LT.ZBEST-.000001) THEN
          DO I = 1,RO
            RBEST(I) = RMEM(I)
          END DO
          DO J = 1,CO
            CBEST(J) = CMEM(J)
          END DO
          ZBEST = ZMIN
          UNIQUE = 1
          NUMOPT = 1
        ELSEIF(ZMIN.GT.ZBEST-.000001.AND.ZMIN.LT.ZBEST+.000001) THEN
          NUMOPT = NUMOPT+1
        END IF
      CALL CPU_TIME(TIMEB)
      IF(TIMEB-TIMEA.LT.TLIMIT) GO TO 1000
      VAF = 1-ZBEST/ASSE
      WRITE(2,22) VAF,NREPS
      DO I = 1,RO
        WRITE(3,*) RBEST(I)
      END DO
      DO I = 1,CO
        WRITE(4,*) CBEST(I)
      END DO
  22  FORMAT(' VAF = ',F12.8,' RESTARTS = ',i10)
      END
