      PROGRAM VNSGSBT2B
      IMPLICIT INTEGER(A-Z)
      REAL*8 TIMEA,TIMEB,TLIMIT,OBJVAL,ZH,YMIN,YPERT,YMAX,
     1       YSTEP
      REAL S1,S2
      INTEGER, ALLOCATABLE :: RCT(:)
      INTEGER, ALLOCATABLE :: CCT(:)
      INTEGER, ALLOCATABLE :: P1(:,:)
      INTEGER, ALLOCATABLE :: P2(:,:)     
      INTEGER, ALLOCATABLE :: Q1(:,:)
      INTEGER, ALLOCATABLE :: Q2(:,:)
      INTEGER, ALLOCATABLE :: PCIR(:)
      INTEGER, ALLOCATABLE :: PRIC(:)
      INTEGER, ALLOCATABLE :: E(:)
      INTEGER, ALLOCATABLE :: EB(:)
      INTEGER, ALLOCATABLE :: M(:)
      INTEGER, ALLOCATABLE :: MB(:)
      INTEGER, ALLOCATABLE :: S(:)
C
C #########################################################################
C  JANUARY 2019 - FORTRAN IMPLEMENTATION OF A VARIABLE NEIGHBORHOOD 
C                 SEARCH HEURISTIC FOR GENERALIZED STRUCTURAL BALANCE
C                 BLOCK PLACEMENTS ARE ASSUMED **UNKNOWN**
C
C     ####### SPARSE INPUT - READS EDGES AND WEIGHTS #######
C     SAME AS VNSGSTB2, EXCEPT THAT IT ASSUMES VERTEX NUMBERING STARTS
C     AT 0 (AS IN THE LEVORATO ET AL. 2017 DATA SETS)
C
C  INPUT FILES:
C      1) PARSET -  SINGLE ROW PARAMETER FILE CONTAINING THE NUMBER OF 
C                   OBJECTS (N), NUMBER OF CLUSTERS (C), NUMBER OF
C                   EDGES (EDGE) AND TIME LIMIT (TL)
C      2) DATAFIL - AN EDGE-BY-3 MATRIX: THE THREE ELEMENTS IN EACH ROW:
C                   (1) START EDGE, (2) END EDGE, AND (3) SIGNED EDGE
C                   WEIGHT (WHICH MUST BE INTEGER)
C  OUTPUT FILES:
C      3) OUTPT   - A SUMMARY OUTPUT FILE THAT REPORTS THE NUMBER OF
C                   INCONSISTENCIES FOR THE RELOCATION PHASE (RH),
C                   FINAL VNS INCONSISTENCIES (OBJVAL), NUMBER OF RESTARTS,
C                   AND MAXIMUM ROW/COLUMN DENSITY
C      4) PARTITION - CLUSTER ASSIGNMENTS OF OBJECTS
C #######################################################################
C
      OPEN(1, FILE='PARSET')
      OPEN(2, FILE='DATAFIL')
      OPEN(3, FILE='OUTPT')
      OPEN(4, FILE='PARTITION')
      READ(1,*) N, C, EDGE, TLIMIT
      ALLOCATE (RCT(N),CCT(N))
      DO I = 1,N
        RCT(I) = 0
        CCT(I) = 0
      END DO
      DO 41 I = 1, EDGE
        READ(2,*) I1, I2, IDX
        if(i1.eq.i2) go to 41
        CCT(I2+1) = CCT(I2+1) + 1
        RCT(I1+1) = RCT(I1+1) + 1
 41   continue
      DENMAX = 0
      DO I = 1,N
        IF(CCT(I).GT.DENMAX) DENMAX = CCT(I)
        IF(RCT(I).GT.DENMAX) DENMAX = RCT(I)
      END DO
      DENMAX = DENMAX+1
      ALLOCATE (E(N),EB(N),M(C),MB(C),S(C))
      ALLOCATE (P1(N,DENMAX),P2(N,DENMAX),Q1(N,DENMAX),Q2(N,DENMAX))
      ALLOCATE (PRIC(N),PCIR(N))
      REWIND 2
      icnt = 0
      DO I = 1,N
        PRIC(I) = 0
        PCIR(I) = 0
        DO J = 1,DENMAX
          P1(I,J) = 0
          P2(I,J) = 0
          Q1(I,J) = 0
          Q2(I,J) = 0
        END DO
      END DO
      DO 40 I = 1, EDGE
        READ(2,*) I1, I2, IDX
        I1 = I1 + 1
        I2 = I2 + 1
        if(i1.eq.i2) go to 40
        if(i1.eq.i2) icnt = icnt + idx
        PRIC(I2) = PRIC(I2) + 1
        PCIR(I1) = PCIR(I1) + 1
        P1(I1,PCIR(I1)) = I2
        P2(I2,PRIC(I2)) = I1
        Q1(I1,PCIR(I1)) = IDX
        Q2(I2,PRIC(I2)) = IDX
 40   continue
      CALL CPU_TIME(TIMEA)  
      DO I = 1,N
        EB(I) = 0
      END DO
      ZBEST = 9999999
      NREPS = 0
 6000 NREPS = NREPS + 1
      DO K = 1,C
        M(K) = 0
        S(K) = 0
      END DO
      DO I = 1,N
        E(I) = 0
      END DO
      DO K = 1,C
c  98    CALL RANDOM_NUMBER(S1)
  98    S1 = RAND()
        ISEL = S1 * FLOAT(N) + 1.
        IF(ISEL.GT.N) ISEL = N
        IF(PCIR(ISEL).EQ.0) GO TO 98
        DO J = 1,K-1
          IF(PCIR(S(J)).EQ.PCIR(ISEL)) GO TO 98
        END DO
        S(K) = ISEL
      END DO
      DO I = 1,N
        IMAX = -99999
        DO K = 1,C
          ISUM = 0
          DO J = 1,PCIR(I)
            DO L = 1,PCIR(S(K))
              IF(P1(I,J).EQ.P1(S(K),L)) THEN
                ISUM = ISUM + Q1(I,J)*Q1(S(K),L)
              END IF
            END DO
          END DO
          DO J = 1,PRIC(I)
            DO L = 1,PRIC(S(K))
              IF(P2(I,J).EQ.P2(S(K),L)) THEN
                ISUM = ISUM + Q2(I,J)*Q2(S(K),L)
              END IF
            END DO
          END DO
          IF(ISUM.GT.IMAX) THEN
            IMAX = ISUM
            KSEL = K
          END IF
        END DO
        E(I) = KSEL
        M(KSEL) = M(KSEL) + 1
      END DO  
      Z = 0
      DO I = 1,N
        K1 = E(I)
        DO J = 1,PCIR(I)
          K2 = E(P1(I,J))
          IF(K1.NE.K2.AND.Q1(I,J).GT.0) Z=Z+Q1(I,J)
          IF(K1.EQ.K2.AND.Q1(I,J).LT.0) Z=Z-Q1(I,J)
        END DO
      END DO
 1000 IFLAG = 0
      DO 100 I = 1,N
        IF(PRIC(I).EQ.0.AND.PCIR(I).EQ.0) GO TO 100
        KI = E(I)
        IF(M(KI).EQ.1) GO TO 100
        DO 120 K = 1,C
          IF(K.EQ.KI) GO TO 120
          DELTA = 0
          DO J = 1,PCIR(I)
            KJ = E(P1(I,J))
            IF(KI.EQ.KJ) DELTA = DELTA + Q1(I,J)
            IF(K.EQ.KJ)  DELTA = DELTA - Q1(I,J)
          END DO
          DO J = 1,PRIC(I)
            KJ = E(P2(I,J))
            IF(KI.EQ.KJ) DELTA = DELTA + Q2(I,J)
            IF(K.EQ.KJ)  DELTA = DELTA - Q2(I,J)
          END DO
          IF(DELTA.LT.0) THEN
            Z = Z + DELTA
            IFLAG = 1
            E(I) = K
            M(K) = M(K) + 1
            M(KI) = M(KI) - 1
            KI = K
          END IF
 120    CONTINUE
 100  CONTINUE
      IF(IFLAG.EQ.1) GO TO 1000
      IF(Z.LT.ZBEST) THEN
        ZBEST = Z
        DO I = 1,N
          EB(I) = E(I)
        END DO
        DO K = 1,C
          MB(K) = M(K)
        END DO
      END IF
      CALL CPU_TIME(TIMEB)
C      IF(TIMEB-TIMEA.LT.TLIMIT) GO TO 6000
      IF(TIMEB-TIMEA.LT.TLIMIT/2) GO TO 6000
C
C ############################################
C   BEGIN THE VARIABLE NEIGHBORHOOD SEARCH
C ############################################
C
      CALL SRAND(3847)
      YMIN = .005 ! .05
      YSTEP = .005 ! .01
      YMAX = .20   ! .50
      YPERT = YMIN
      DO I = 1,N
        E(I) = EB(I)
      END DO
      DO K = 1,C
        M(K) = MB(K)
      END DO
      Z = ZBEST
      ZH = ZBEST
C
C ######### PERTURB ########
C
 150  DO 151 I = 1,N
        IF(PRIC(I).EQ.0.AND.PCIR(I).EQ.0) GO TO 151
        KI = E(I)
        IF(M(KI).GT.1) THEN
C          CALL RANDOM_NUMBER(S1)
          S1 = RAND()
          IF(S1.LT.YPERT) THEN
C 199        CALL RANDOM_NUMBER(S2)
  199       S2 = RAND()
            KSEL = 1+S2*FLOAT(C)
            IF(KSEL.EQ.KI) KSEL = KSEL + 1
            IF(KSEL.GT.C) KSEL = 1
            DELTA = 0
            DO J = 1,PCIR(I)
              KJ = E(P1(I,J))
              IF(KI.EQ.KJ)   DELTA = DELTA + Q1(I,J)
              IF(KSEL.EQ.KJ) DELTA = DELTA - Q1(I,J)
            END DO
            DO J = 1,PRIC(I)
              KJ = E(P2(I,J))
              IF(KI.EQ.KJ)   DELTA = DELTA + Q2(I,J)
              IF(KSEL.EQ.KJ) DELTA = DELTA - Q2(I,J)
            END DO
            Z = Z + DELTA
            E(I) = KSEL
            M(KSEL) = M(KSEL) + 1
            M(KI) = M(KI) - 1
          END IF
        END IF
 151  CONTINUE
C
C ######### REOPTIMIZE ########
C
 1001 IFLAG = 0
      DO 101 I = 1,N
        IF(PRIC(I).EQ.0.AND.PCIR(I).EQ.0) GO TO 101
        KI = E(I)
        IF(M(KI).EQ.1) GO TO 101
        DO 121 K = 1,C
          IF(K.EQ.KI) GO TO 121
          DELTA = 0
          DO J = 1,PCIR(I)
            KJ = E(P1(I,J))
            IF(KI.EQ.KJ)  DELTA = DELTA + Q1(I,J)
            IF (K.EQ.KJ)  DELTA = DELTA - Q1(I,J)
          END DO
          DO J = 1,PRIC(I)
            KJ = E(P2(I,J))
            IF(KI.EQ.KJ)  DELTA = DELTA + Q2(I,J)
            IF(K.EQ.KJ)   DELTA = DELTA - Q2(I,J)
          END DO
          IF(DELTA.LT.0) THEN
            Z = Z + DELTA
            IFLAG = 1
            E(I) = K
            M(K) = M(K) + 1
            M(KI) = M(KI) - 1
            KI = K
          END IF
 121    CONTINUE
 101  CONTINUE
      IF(IFLAG.EQ.1) GO TO 1001
      IF(Z.LT.ZBEST) THEN
        ZBEST = Z
        DO I = 1,N
          EB(I) = E(I)
        END DO
        YPERT = YMIN
      ELSE
        YPERT = YPERT + YSTEP
        IF(YPERT.GT.YMAX) YPERT = YMIN
      END IF
      CALL CPU_TIME(TIMEB)
      IF(TIMEB-TIMEA.LT.TLIMIT) THEN
        DO I = 1,N
          E(I) = EB(I)
        END DO
        DO K = 1,C
          M(K) = MB(K)
        END DO
        Z = ZBEST
        GO TO 150     
      END IF
      OBJVAL = ZBEST
      DO I = 1,N
        WRITE(4,*) EB(I)
      END DO
      WRITE(3,413) DENMAX
      WRITE(3,416) NREPS
      WRITE(3,414) ZH
      WRITE(3,415) OBJVAL
  413 FORMAT(' MAXIMUM NONZEROS IN ROWS OR COLUMNS ',I8)
 414  FORMAT(' RELOCATION HEURISTIC INCONSITENCIES ',F8.1)
 415  FORMAT(' VARIABLE NEIG SEARCH INCONSITENCIES ',F8.1)
  416 FORMAT(' RELOCATION HEURISTIC TOTAL RESTARTS ',I8)
 889  continue
      END
